/* ./amel/stimulation_multiscale.cpp 
**********************************
Copyright EPFL and INRIA

author : Guillaume ANCIAUX (guillaume.anciaux@epfl.fr, g.anciaux@laposte.net)

The LibMultiScale is a C++ parallel framework for the multiscale
coupling methods dedicated to material simulations. This framework
provides an API which makes it possible to program coupled simulations
and integration of already existing codes.

This Project is done in a collaboration between 
EPFL within ENAC-LSMS (http://lsms.epfl.ch/) and 
INRIA Bordeaux, ScAlApplix (http://www.labri.fr/projet/scalapplix/). 

This software is governed by the CeCILL-C license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL-C
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL-C license and that you accept its terms.
***********************************/

#define NEED_STAMP
#define NEED_LAMMPS
#define NEED_MD1D
#define NEED_LIBMESH
#define LOCAL_MODULE MOD_STIMULATION

#include "../common/lm_common.h"
#include "../stimulation/stimulation.h"
#include "stimulation_multiscale.h"

#define INSTANCIATIONA1(template_class,stimulation_class) if (dynamic_cast<template_class*>(domA)){ \
      if (Dim == 1)stimulators.push_back(new stimulation_class<template_class,1>(*domA));}

#define INSTANCIATIONA2(template_class,stimulation_class) if (dynamic_cast<template_class*>(domA)){ \
      if (Dim == 2)stimulators.push_back(new stimulation_class<template_class,2>(*domA));}

#define INSTANCIATIONA3(template_class,stimulation_class) if (dynamic_cast<template_class*>(domA)){ \
      if (Dim == 3)stimulators.push_back(new stimulation_class<template_class,3>(*domA));} 


StimulationMultiScale::StimulationMultiScale(DomainMultiScale & d,unsigned int dim)
  :dom(d),Dim(dim){}

std::vector<int> StimulationMultiScale::CreateStimulation(int type// ,double intensity,int bID,unsigned int frequency
							  ){
  std::vector<int> list;

  for (unsigned int i = 0 ; i < dom.nb_models(); ++i)
    {
      int mod_type = dom.modelByNumber(i)->GetType();
      if (mod_type == ATOMTYPE)
	list.push_back(CreateStimulationAtoms(type,dom.modelByNumber(i)// ,intensity,bID,frequency
					      ));
      if (mod_type == CONTINUTYPE)
	list.push_back(CreateStimulationContinu(type,dom.modelByNumber(i)// ,intensity,bID,frequency
						));
    }

  return list;
}

int StimulationMultiScale::CreateStimulationAtoms(int type,DomainInterface * domA){

  switch(type){
  case DEPUNI:
#ifdef USING_STAMP
    INSTANCIATIONA1(DomainStamp,StimulationDepUni);
#endif
#ifdef USING_LAMMPS
    INSTANCIATIONA2(DomainLammps,StimulationDepUni);
    INSTANCIATIONA3(DomainLammps,StimulationDepUni);
#endif
#ifdef USING_MD1D
    INSTANCIATIONA1(DomainMD1D,StimulationDepUni);
#endif
    break;
  case VELUNI:
#ifdef USING_LAMMPS
    INSTANCIATIONA2(DomainLammps,StimulationVelUni);
    INSTANCIATIONA3(DomainLammps,StimulationVelUni);
#endif
    break;
  case WALL:
#ifdef USING_LAMMPS
    INSTANCIATIONA3(DomainLammps,StimulationWall);
#endif
    break;
  case MYSHEAR:
#ifdef USING_LAMMPS
    INSTANCIATIONA3(DomainLammps,StimulationShear);
#endif
    break;
  case SINUSOIDE:
#ifdef USING_STAMP
    INSTANCIATIONA123(DomainStamp,StimulationSinusoide);
#endif
#ifdef USING_LAMMPS
    INSTANCIATIONA2(DomainLammps,StimulationSinusoide);
    INSTANCIATIONA3(DomainLammps,StimulationSinusoide);
#endif
#ifdef USING_MD1D
    INSTANCIATIONA1(DomainMD1D,StimulationSinusoide);
#endif
    break;
  case RAZ:
#ifdef USING_STAMP
    INSTANCIATIONA123(DomainStamp,StimulationRAZ);
#endif
#ifdef USING_LAMMPS
    INSTANCIATIONA2(DomainLammps,StimulationRAZ);
    INSTANCIATIONA3(DomainLammps,StimulationRAZ);
#endif
#ifdef USING_MD1D
    INSTANCIATIONA1(DomainMD1D,StimulationRAZ);
#endif
    break;
  case DEPGAUSS:
#ifdef USING_STAMP
    INSTANCIATIONA123(DomainStamp,StimulationDepGauss);
#endif
#ifdef USING_LAMMPS
    INSTANCIATIONA2(DomainLammps,StimulationDepGauss);
    INSTANCIATIONA3(DomainLammps,StimulationDepGauss);
#endif
#ifdef USING_MD1D
    INSTANCIATIONA1(DomainMD1D,StimulationDepGauss);
#endif
    break;
  case VELGAUSS:
#ifdef USING_STAMP
    INSTANCIATIONA123(DomainStamp,StimulationVelGauss);
#endif
#ifdef USING_LAMMPS
    INSTANCIATIONA2(DomainLammps,StimulationVelGauss);
    INSTANCIATIONA3(DomainLammps,StimulationVelGauss);
#endif
#ifdef USING_MD1D
    INSTANCIATIONA1(DomainMD1D,StimulationVelGauss);
#endif
    break;
  case MODULATE:
#ifdef USING_STAMP
    INSTANCIATIONA123(DomainStamp,StimulationModulation);
#endif
#ifdef USING_LAMMPS
    INSTANCIATIONA2(DomainLammps,StimulationModulation);
    INSTANCIATIONA3(DomainLammps,StimulationModulation);
#endif
#ifdef USING_MD1D
    INSTANCIATIONA1(DomainMD1D,StimulationModulation);
#endif
    break;
  case MATIERELANCEE:
#ifdef USING_STAMP
    INSTANCIATIONA123(DomainStamp,StimulationMatiereLancee);
#endif
#ifdef USING_LAMMPS
    INSTANCIATIONA2(DomainLammps,StimulationMatiereLancee);
    INSTANCIATIONA3(DomainLammps,StimulationMatiereLancee);
#endif
#ifdef USING_MD1D
    INSTANCIATIONA1(DomainMD1D,StimulationMatiereLancee);
#endif
    break;
  case STABILISE:
#ifdef USING_STAMP
    INSTANCIATIONA123(DomainStamp,StimulationStabilise);
#endif
#ifdef USING_LAMMPS
    INSTANCIATIONA2(DomainLammps,StimulationStabilise);
    INSTANCIATIONA3(DomainLammps,StimulationStabilise);
#endif
#ifdef USING_MD1D
    INSTANCIATIONA1(DomainMD1D,StimulationStabilise);
#endif
    break;
  case IMPULSE:
#ifdef USING_STAMP
    INSTANCIATIONA123(DomainStamp,StimulationImpulse);
#endif
#ifdef USING_LAMMPS
    INSTANCIATIONA2(DomainLammps,StimulationImpulse);
    INSTANCIATIONA3(DomainLammps,StimulationImpulse);
#endif
#ifdef USING_MD1D
    INSTANCIATIONA1(DomainMD1D,StimulationImpulse);
#endif
    break;
  case DISLOEDGE:
#ifdef USING_STAMP
    INSTANCIATIONA123(DomainStamp,StimulationDisloEdge);
#endif
#ifdef USING_LAMMPS
    INSTANCIATIONA2(DomainLammps,StimulationDisloEdge);
    INSTANCIATIONA3(DomainLammps,StimulationDisloEdge);
#endif
#ifdef USING_MD1D
    INSTANCIATIONA1(DomainMD1D,StimulationDisloEdge);
#endif
    break;
  case DISLOSCREW:
#ifdef USING_STAMP
    INSTANCIATIONA123(DomainStamp,StimulationDisloScrew);
#endif
#ifdef USING_LAMMPS
    INSTANCIATIONA2(DomainLammps,StimulationDisloScrew);
    INSTANCIATIONA3(DomainLammps,StimulationDisloScrew);
#endif
#ifdef USING_MD1D
    INSTANCIATIONA1(DomainMD1D,StimulationDisloScrew);
#endif
    break;
  case DEPPLAN:
#ifdef USING_LAMMPS
    INSTANCIATIONA3(DomainLammps,StimulationDepPlan);
#endif
    break;
  }
  if (stimulators.size() == 0 || !stimulators[stimulators.size()-1]){
    FATAL("This combination of model (" << type << ") and dimension (" << Dim << ") does not exist" << Dim);
  }

  stimulators[stimulators.size()-1]->frequency = frequency;
  return (stimulators.size()-1);
}

#define INSTANCIATIONM1(type,stimulateur_class)  if (dynamic_cast<type*>(domC) && (Dim == 1)) \
   	           stimulators.push_back(new stimulateur_class<type,1>(*domC)); 

#define INSTANCIATIONM2(type,stimulateur_class)  if (dynamic_cast<type*>(domC) && (Dim == 2)) \
                   stimulators.push_back(new stimulateur_class<type,2>(*domC)); 

#define INSTANCIATIONM3(type,stimulateur_class)  if (dynamic_cast<type*>(domC) && (Dim == 3)) \
                   stimulators.push_back(new stimulateur_class<type,3>(*domC));


int StimulationMultiScale::CreateStimulationContinu(int type,DomainInterface * domC){

  switch(type){
  case DEPUNI:
#ifdef USING_LIBMESH
    INSTANCIATIONM1(DomainLibMesh<1>,StimulationDepUni);
    INSTANCIATIONM2(DomainLibMesh<2>,StimulationDepUni);
    INSTANCIATIONM3(DomainLibMesh<3>,StimulationDepUni);
#endif
#ifdef USING_SIMULPACK
    INSTANCIATIONM3(DomainSimulPack<3>,StimulationDepUni);
#endif
#ifdef USING_MECA1D
    INSTANCIATIONM1(DomainMeca1D,StimulationDepUni);
#endif
    break;
  case WALL:
#ifdef USING_SIMULPACK
    INSTANCIATIONM3(DomainSimulPack<3>,StimulationWall);
#endif
    break;
  case MYSHEAR:
#ifdef USING_SIMULPACK
    INSTANCIATIONM3(DomainSimulPack<3>,StimulationShear);
#endif
    break;
  case SINUSOIDE:
#ifdef USING_LIBMESH
    INSTANCIATIONM1(DomainLibMesh<1>,StimulationSinusoide);
    INSTANCIATIONM2(DomainLibMesh<2>,StimulationSinusoide);
    INSTANCIATIONM3(DomainLibMesh<3>,StimulationSinusoide);
#endif
#ifdef USING_MECA1D
    INSTANCIATIONM1(DomainMeca1D,StimulationSinusoide);
#endif
    break;
  case RAZ:
#ifdef USING_LIBMESH
    INSTANCIATIONM1(DomainLibMesh<1>,StimulationRAZ);
    INSTANCIATIONM2(DomainLibMesh<2>,StimulationRAZ);
    INSTANCIATIONM3(DomainLibMesh<3>,StimulationRAZ);
#endif
#ifdef USING_MECA1D
    INSTANCIATIONM1(DomainMeca1D,StimulationRAZ);
#endif
#ifdef USING_SIMULPACK
    INSTANCIATIONM3(DomainSimulPack<3>,StimulationRAZ);
#endif
    break;
  case DEPGAUSS:
#ifdef USING_LIBMESH
    INSTANCIATIONM1(DomainLibMesh<1>,StimulationDepGauss);
    INSTANCIATIONM2(DomainLibMesh<2>,StimulationDepGauss);
    INSTANCIATIONM3(DomainLibMesh<3>,StimulationDepGauss);
#endif
#ifdef USING_SIMULPACK
    INSTANCIATIONM2(DomainSimulPack<2>,StimulationDepGauss);
    INSTANCIATIONM3(DomainSimulPack<3>,StimulationDepGauss);
#endif
#ifdef USING_MECA1D
    INSTANCIATIONM1(DomainMeca1D,StimulationDepGauss);
#endif
    break;
  case MODULATE:
#ifdef USING_LIBMESH
    INSTANCIATIONM1(DomainLibMesh<1>,StimulationModulation);
    INSTANCIATIONM2(DomainLibMesh<2>,StimulationModulation);
    INSTANCIATIONM3(DomainLibMesh<3>,StimulationModulation);
#endif
#ifdef USING_MECA1D
    INSTANCIATIONM1(DomainMeca1D,StimulationModulation);
#endif
    break;
  case MATIERELANCEE:
#ifdef USING_LIBMESH
    INSTANCIATIONM1(DomainLibMesh<1>,StimulationMatiereLancee);
    INSTANCIATIONM2(DomainLibMesh<2>,StimulationMatiereLancee);
    INSTANCIATIONM3(DomainLibMesh<3>,StimulationMatiereLancee);
#endif
#ifdef USING_MECA1D
    INSTANCIATIONM1(DomainMeca1D,StimulationMatiereLancee);
#endif
    break;
  case STABILISE:
#ifdef USING_LIBMESH
    INSTANCIATIONM1(DomainLibMesh<1>,StimulationStabilise);
    INSTANCIATIONM2(DomainLibMesh<2>,StimulationStabilise);
    INSTANCIATIONM3(DomainLibMesh<3>,StimulationStabilise);
#endif
#ifdef USING_MECA1D
    INSTANCIATIONM1(DomainMeca1D,StimulationStabilise);
#endif
#ifdef USING_SIMULPACK
    INSTANCIATIONM3(DomainSimulPack<3>,StimulationStabilise);
#endif
    break;
  case IMPULSE:
#ifdef USING_LIBMESH
    INSTANCIATIONM1(DomainLibMesh<1>,StimulationImpulse);
    INSTANCIATIONM2(DomainLibMesh<2>,StimulationImpulse);
    INSTANCIATIONM3(DomainLibMesh<3>,StimulationImpulse);
#endif
#ifdef USING_MECA1D
    INSTANCIATIONM1(DomainMeca1D,StimulationImpulse);
#endif
#ifdef USING_SIMULPACK
    INSTANCIATIONM3(DomainSimulPack<3>,StimulationImpulse);
    INSTANCIATIONM2(DomainSimulPack<2>,StimulationImpulse);
#endif
    break;
  case DISLOEDGE:
#ifdef USING_LIBMESH
    INSTANCIATIONM1(DomainLibMesh<1>,StimulationDisloEdge);
    INSTANCIATIONM2(DomainLibMesh<2>,StimulationDisloEdge);
    INSTANCIATIONM3(DomainLibMesh<3>,StimulationDisloEdge);
#endif
#ifdef USING_SIMULPACK
    INSTANCIATIONM3(DomainSimulPack<3>,StimulationDisloEdge);
#endif
#ifdef USING_MECA1D
    INSTANCIATIONM1(DomainMeca1D,StimulationDisloEdge);
#endif
    break;
  case DISLOSCREW:
#ifdef USING_LIBMESH
    INSTANCIATIONM1(DomainLibMesh<1>,StimulationDisloScrew);
    INSTANCIATIONM2(DomainLibMesh<2>,StimulationDisloScrew);
    INSTANCIATIONM3(DomainLibMesh<3>,StimulationDisloScrew);
#endif
#ifdef USING_SIMULPACK
    INSTANCIATIONM3(DomainSimulPack<3>,StimulationDisloScrew);
#endif
#ifdef USING_MECA1D
    INSTANCIATIONM1(DomainMeca1D,StimulationDisloScrew);
#endif
    break;
  case DEPPLAN:
#ifdef USING_LIBMESH
    INSTANCIATIONM1(DomainLibMesh<1>,StimulationDepPlan);
    INSTANCIATIONM2(DomainLibMesh<2>,StimulationDepPlan);
    INSTANCIATIONM3(DomainLibMesh<3>,StimulationDepPlan);
#endif
#ifdef USING_MECA1D
    INSTANCIATIONM1(DomainMeca1D,StimulationDepPlan);
#endif
#ifdef USING_SIMULPACK
    INSTANCIATIONM3(DomainSimulPack<3>,StimulationDepPlan);
#endif
    break;
  }
  
  if (stimulators.size() == 0 || !stimulators[stimulators.size()-1]){
    FATAL("This combination of model (" << type << ") and dimension (" << Dim << ") does not exist" << Dim);}

  DUMP("number of registered domains for stimulation : " << stimulators.size(),DBG_INFO);

  stimulators[stimulators.size()-1]->frequency = frequency;
  return (stimulators.size() - 1);
}


void StimulationMultiScale::Stimulate(int stage)
{
  for (unsigned int i = 0 ; i < stimulators.size(); ++i){
    DUMP("for test " << i << " " << stimulators[i]->frequency << " " << stim_step,DBG_INFO);
    if (stimulators[i]->end_step >= current_step && stimulators[i]->start_step <= current_step)
      if (stimulators[i]->frequency != -1 && current_step%stimulators[i]->frequency == 0){
	DUMP("proceed stimulation of stimulator " << i << " current_step = " << current_step << " start " 
	     << stimulators[i]->start_step << " end " 
	     << stimulators[i]->end_step,DBG_INFO);
	stimulators[i]->Stimulate(stage);      
	++stimulators[i]->stim_step;
      }
  }
  ++stim_step;
}


StimulationInterface * StimulationMultiScale::GetStimulation(unsigned i)
{
  return stimulators[i];
}
  

StimulationMultiScale::~StimulationMultiScale(){

  for (unsigned int i = 0 ; i < stimulators.size(); ++i)
    delete stimulators[i];

}

int StimulationMultiScale::ParseLine(char * line){
  char mot[256];
  int stim_type=-1;

  line+=Parser::strNext(mot,line);

  if(!strcmp(mot,"GAUSS")) stim_type = DEPGAUSS;
  else if(!strcmp(mot,"PLAN")) stim_type = DEPPLAN;
  else if(!strcmp(mot,"IMPULSE")) stim_type = IMPULSE;
  else if(!strcmp(mot,"DISLOEDGE")) stim_type = DISLOEDGE;
  else if(!strcmp(mot,"DISLOSCREW")) stim_type = DISLOSCREW;
  else if(!strcmp(mot,"MODULATE")) stim_type = MODULATE;
  else if(!strcmp(mot,"DEPUNI")) stim_type = DEPUNI;
  else if(!strcmp(mot,"VELUNI")) stim_type = VELUNI;
  else if(!strcmp(mot,"SIN")) stim_type = SINUSOIDE;
  else if(!strcmp(mot,"RAZ")) stim_type = RAZ;
  else if(!strcmp(mot,"MATIERELANCEE")) stim_type = MATIERELANCEE;
  else if(!strcmp(mot,"STABILISE")) stim_type = STABILISE;
  else if(!strcmp(mot,"DISLOEDGEBORDERS")) stim_type = DISLOEDGE;
  else if(!strcmp(mot,"WALL")) stim_type = WALL;
  else if(!strcmp(mot,"SHEAR")) stim_type = MYSHEAR;
  else FATAL("Unknown stimulator type " << mot);

  std::vector<int> stim_indexes = CreateStimulation(stim_type);

  int lu = Parser::strNext(mot,line);  
  while(lu){
    line += lu;
    char key[50];
    double value;
    strcpy(key,mot);
    line += Parser::parseDouble(value,line);    
    DUMP("Parsing STIMULATION COMMAND " << key << " : " << value,DBG_INFO_STARTUP);
    for (unsigned k = 0 ; k < stim_indexes.size() ; ++k){
      int parsed = stimulators[stim_indexes[k]]->SetParam(key,value);
      if (!parsed) FATAL("parameter " << key << " was not parsed !! check config file for a syntax problem");
    }
    lu = Parser::strNext(mot,line);
  }
  return true;
}

int StimulationMultiScale::SetParam(char* key,double value){
  int parsed = true;
  for (unsigned int i = 0 ; i < stimulators.size(); ++i)
    parsed &= stimulators[i]->SetParam(key,value);

  return parsed;
}
